""" Generators for parameter functions """

from dune.perftool.generation import (class_basename,
                                      class_member,
                                      constructor_parameter,
                                      generator_factory,
                                      get_backend,
                                      initializer_list,
                                      kernel_cached,
                                      preamble,
                                      temporary_variable
                                      )
from dune.perftool.pdelab.geometry import (name_cell,
                                           name_dimension,
                                           name_intersection,
                                           )
from dune.perftool.pdelab.quadrature import (pymbolic_quadrature_position,
                                             pymbolic_quadrature_position_in_cell,
                                             quadrature_preamble,
                                             )
from dune.perftool.tools import get_pymbolic_basename
from dune.perftool.cgen.clazz import AccessModifier
from dune.perftool.pdelab.localoperator import (class_type_from_cache,
                                                localoperator_basename,
                                                )


@class_basename("parameterclass")
def parameterclass_basename(formdata, data):
    lopbase = localoperator_basename(formdata, data)
    return "{}Params".format(lopbase)


@class_member("operator")
def define_parameterclass(name):
    _, t = class_type_from_cache("parameterclass")
    constructor_parameter("const {}&".format(t), name + "_", classtag="operator")
    initializer_list(name, [name + "_"], classtag="operator")
    return "const {}& {};".format(t, name)


def name_paramclass():
    define_parameterclass("param")
    return "param"


@class_member(classtag="parameterclass", access=AccessModifier.PRIVATE)
def define_time(name):
    initializer_list(name, ["0.0"], classtag="parameterclass")
    return "double {};".format(name)


def name_time():
    define_time("t")
    return "t"


@class_member("parameterclass", access=AccessModifier.PUBLIC)
def define_set_time_method():
    time_name = name_time()
    # TODO double?
    result = ["// Set time in instationary case",
              "void setTime (double t_)",
              "{",
              "  {} = t_;".format(time_name),
              "}"
              ]

    return result


def component_to_tree_path(element, component):
    subel = element.extract_subelement_component(component)

    def _flatten(x):
        if isinstance(x, tuple):
            return '_'.join(_flatten(i) for i in x if i != ())
        else:
            return str(x)

    return _flatten(subel)


@class_member("parameterclass", access=AccessModifier.PUBLIC)
def define_parameter_function_class_member(name, expr, baset, shape, cell):
    t = construct_nested_fieldvector(baset, shape)

    geot = "E" if cell else "I"
    geo = geot.lower()
    result = ["template<typename {}, typename X>".format(geot),
              "{} {}(const {}& {}, const X& local) const".format(t, name, geot, geo),
              "{",
              ]

    # In the case of a non-scalar parameter function, recurse into leafs
    if expr.element.value_shape():
        # Check that this is a VectorElement, as I have no idea how a parameter function
        # over a non-vector mixed element should be well-defined in PDELab.
        from ufl import VectorElement
        assert isinstance(expr.element, VectorElement)

        result.append("  {} result(0.0);".format(t))

        from dune.perftool.ufl.execution import split_expression
        for i, subexpr in enumerate(split_expression(expr)):
            child_name = "{}_{}".format(name, component_to_tree_path(expr.element, i))
            result.append("  result[{}] = {}({}, local);".format(i, child_name, geo))
            define_parameter_function_class_member(child_name, subexpr, baset, shape[1:], cell)

        result.append("  return result;")

    else:
        # Evaluate a scalar parameter function
        if expr.is_global:
            result.append("  auto x = {}.geometry().global(local);".format(geo))
        else:
            result.append("  auto x = local;")

        result.append("  " + expr.c_expr[0])

    result.append("}")

    return result


@preamble
def evaluate_cellwise_constant_parameter_function(name, restriction):
    param = name_paramclass()
    entity = name_cell(restriction)
    from dune.perftool.pdelab.geometry import name_localcenter
    pos = name_localcenter()

    from dune.perftool.generation.loopy import valuearg
    import numpy
    valuearg(name, dtype=numpy.float64)

    return 'auto {} = {}.{}({}, {});'.format(name,
                                             name_paramclass(),
                                             name,
                                             entity,
                                             pos,
                                             )


@preamble
def evaluate_intersectionwise_constant_parameter_function(name):
    # Check that this is not a volume term, as that would not be well-defined
    from dune.perftool.generation import get_global_context_value
    it = get_global_context_value("integral_type")
    assert it is not 'cell'

    param = name_paramclass()
    intersection = name_intersection()
    pos = name_localcenter()

    from dune.perftool.generation.loopy import valuearg
    import numpy
    valuearg(name, dtype=numpy.float64)

    return 'auto {} = {}.{}({}, {});'.format(name,
                                             name_paramclass(),
                                             name,
                                             intersection,
                                             pos,
                                             )


def evaluate_cell_parameter_function(name, restriction):
    param = name_paramclass()
    entity = name_cell(restriction)
    pos = get_backend(interface="qp_in_cell")(restriction)
    return quadrature_preamble('{} = {}.{}({}, {});'.format(name,
                                                            name_paramclass(),
                                                            name,
                                                            entity,
                                                            str(pos),
                                                            ),
                               assignees=frozenset({name}),
                               read_variables=frozenset({get_pymbolic_basename(pos)}),
                               )


def evaluate_intersection_parameter_function(name):
    # Check that this is not a volume term, as that would not be well-defined
    from dune.perftool.generation import get_global_context_value
    it = get_global_context_value("integral_type")
    assert it is not 'cell'

    param = name_paramclass()
    intersection = name_intersection()
    pos = get_backend("quad_pos")()
    return quadrature_preamble('{} = {}.{}({}, {});'.format(name,
                                                            name_paramclass(),
                                                            name,
                                                            intersection,
                                                            str(pos),
                                                            ),
                               assignees=frozenset({name}),
                               read_variables=frozenset({get_pymbolic_basename(pos)}),
                               )


def construct_nested_fieldvector(t, shape):
    if len(shape) == 0:
        return t
    return 'Dune::FieldVector<{}, {}>'.format(construct_nested_fieldvector(t, shape[1:]), shape[0])


@kernel_cached
def cell_parameter_function(name, expr, restriction, cellwise_constant, t='double'):
    shape = expr.ufl_element().value_shape()
    shape_impl = ('fv',) * len(shape)
    define_parameter_function_class_member(name, expr, t, shape, True)
    if cellwise_constant:
        evaluate_cellwise_constant_parameter_function(name, restriction)
    else:
        temporary_variable(name, shape=shape, shape_impl=shape_impl)
        evaluate_cell_parameter_function(name, restriction)


@kernel_cached
def intersection_parameter_function(name, expr, cellwise_constant, t='double'):
    shape = expr.ufl_element().value_shape()
    shape_impl = ('fv',) * len(shape)
    define_parameter_function_class_member(name, expr, t, shape, False)
    if cellwise_constant:
        evaluate_intersectionwise_constant_parameter_function(name)
    else:
        temporary_variable(name, shape=shape, shape_impl=shape_impl)
        evaluate_intersection_parameter_function(name)
